import configparser
import os

''' 
method_dxdoiorg                         It sets which method is used by the library pdf2doi when querying dx.doi.org to retrieve the bibtex info
                                        Two possible values are 'text/bibliography; style=bibtex' , 'application/x-bibtex' and
                                        'application/citeproc+json'                             
'''


class config():
    __params={'verbose'   :   True,
            'separator' : os.path.sep,
            'method_dxdoiorg' : 'application/citeproc+json',
            }
    __setters = __params.keys()


    @staticmethod
    def update_params(new_params):
        config.__params.update(new_params)

    @staticmethod
    def get(name):
        return config.__params[name]

    @staticmethod
    def set(name, value):
        if name in config.__setters:
             config.__params[name] = value
        else:
            raise NameError("Name not accepted in set() method")

    @staticmethod
    def ReadParamsINIfile():
        '''
        Reads the parameters stored in the file settings.ini, and stores them in the dict self.params
        If the .ini file does not exist, it creates it with the default values.
        '''
        path_current_directory = os.path.dirname(__file__)
        path_config_file = os.path.join(path_current_directory, 'settings.ini')
        if not(os.path.exists(path_config_file)):
            config.WriteParamsINIfile()
        else:
            config_object = configparser.ConfigParser()
            config_object.optionxform = str
            config_object.read(path_config_file)
            config.__params.update(dict(config_object['DEFAULT']))
            config.ConvertParamsToNumb()

    @staticmethod
    def ConvertParamsToNumb():
        for key,val in config.__params.items():
            if isinstance(val, str) and val.isdigit():
                config.__params[key]=int(val)
    @staticmethod
    def print():
        '''
        Prints all settings
        '''
        for key,val in config.__params.items():
            print(key + " : " + str(val))

    @staticmethod
    def WriteParamsINIfile():
        '''
        Writes the parameters currently stored in in the dict self.params into the file settings.ini
        '''
        #params_dict = {key: str(value) for key, value in (config.__params).items()} #Make sure that each value of the dict is converted to a string
        path_current_directory = os.path.dirname(__file__)
        path_config_file = os.path.join(path_current_directory, 'settings.ini')
        config_object = configparser.ConfigParser()
        config_object.optionxform = str
        #print(config_object)
        config_object['DEFAULT'] = config.__params
        with open(path_config_file, 'w') as configfile: #Write them on file
            config_object.write(configfile)


###########################################











import logging

# Setup logging
logger = logging.getLogger("pdf2bib")
logger.setLevel(level=logging.INFO)
if not logger.handlers:
    formatter = logging.Formatter("[pdf2bib]: %(message)s")
    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    logger.addHandler(ch)
logger.propagate = False

from .config import config
config.ReadParamsINIfile()

from .main import pdf2bib
from .bibtex_makers import *
#from .utils_registry import install_right_click, uninstall_right_click





import argparse
import logging
from os import path, listdir
import pdf2bib.bibtex_makers as bibtex_makers
import pdf2bib.config as config
import pdf2doi
#import pyperclip Modules that are commented here are imported later only when needed, to improve start up time

def pdf2bib(target):
    ''' This is the main routine of the library. When the library is used as a command-line tool (via the entry-point "pdf2bib") 
    ''' 

    # Setup logging
    if config.get('verbose'): loglevel = logging.INFO
    else: loglevel = logging.CRITICAL

    logger = logging.getLogger("pdf2bib")
    logger.setLevel(level=loglevel)

    #Check if path is valid
    if not(path.exists(target)):
        logger.error(f"{target} is not a valid path to a file or a directory.")
        return None
      
    #Check if target is a directory
    #If yes, we look for all the .pdf files inside it, and for each of them
    #we call again this function
    if  path.isdir(target):
        logger.info(f"Looking for pdf files in the folder {target}...")
        pdf_files = [f for f in listdir(target) if (f.lower()).endswith('.pdf')]
        numb_files = len(pdf_files)
        
        if numb_files == 0:
            logger.error("No pdf files found in this folder.")
            return None
        
        logger.info(f"Found {numb_files} pdf files.")
        if not(target.endswith(config.get('separator'))): #Make sure the path ends with "\" or "/" (according to the OS)
            target = target + config.get('separator')
            
        papers = [] #For each pdf file in the target folder we will store a dictionary inside this list
        for f in pdf_files:
            logger.info("................") 
            file = target + f
            #For each file we call again this function, but now the input argument target is set to the path of the file
            result = pdf2bib(target=file)
            #logger.info(result['identifier'])
            papers.append(result)

        logger.info("................") 

        return papers
    
    #If target is not a directory, we check that it is an existing file and that it ends with .pdf
    else:
        filename = target
        if not path.exists(filename):
            logger.error(f"'{filename}' is not a valid file.")
            return None   
        if not (filename.lower()).endswith('.pdf'):
            logger.error("The file must have .pdf extension.")
            return None
        logger.info(f"Trying to extract data to generate the BibTeX entry for the file: {filename}")  
        logger.info(f"Calling pdf2doi...") 
        result = pdf2doi.pdf2doi_singlefile(filename)
        if result['identifier'] == None:
            logger.error("It was not possible to find a valid identifier for this file.")
        if not isinstance(result['validation_info'],str):
            logger.error("The validation_info returned by pdf2doi is not a string. It is not possible to extract BibTeX data.")
        logger.info(f"Trying to parse the data obtained by pdf2doi into valid BibTeX data..") 
        metadata = pdf2bib_singlefile(result)
        if metadata:
            result['metadata'] = metadata
            result['bibtex'] = bibtex_makers.make_bibtex(metadata)
            logger.info(f"A valid BibTeX entry was generated:") 
            logger.info(result['bibtex'])
            return result 
        else:
            logger.error("Some error occurred when parsing the raw BibTeX data.")

def pdf2bib_singlefile(result):
    '''
    Parameters
    ----------
    result : dictionary
        dictionary obtained via pdf2doi
    Returns
    -------
    ''' 
    
    if result['identifier_type'] == 'DOI':
        metadata = bibtex_makers.parse_bib_from_dxdoiorg(result['validation_info'], method=pdf2doi.config.get('method_dxdoiorg'))
    if result['identifier_type'] == 'arxiv ID':
        metadata = bibtex_makers.parse_bib_from_exportarxivorg(result['validation_info'])
    
    return metadata



def save_bibtex_entries(filename_bibtex, results, clipboard = False):
    ''' Write all bibtex entries contained in the input list 'results' into a text file with a path specified by filename_bibtex 
        (if filename_bibtex is a valid string) and/or into the clipboard (if clipboard = True).
    
    Parameters
    ----------
    filename_bibtex : string
        Absolute path of the target file. If equal to '' or False, nothing is stored on file.
    results : list of dictionaries
        Each element of the list 'results' describes a .pdf file, and contains the pdf identifier, bibtex entry and other infos.
    clipboard : boolean, optional
        If set to True, the bibtex entries are stored in the clipboard. Default is False.

    Returns
    -------
    None.
    '''
    logger = logging.getLogger("pdf2bib")

    if isinstance(filename_bibtex,str) or clipboard:
        text = ''
        for result in results:
            if isinstance(result['bibtex'],str):
                text = text + result['bibtex'] + "\n\n"

    #If filename_bibtex is a valid string, we create the full path of the file where bibtex entries will be saved
    if isinstance(filename_bibtex,str):
        path_filename_bibtex = path.dirname(results[0]['path']) + config.get('separator') + filename_bibtex
        try:
            with open(path_filename_bibtex, "w", encoding="utf-8") as text_file:
                text_file.write(text) 
                logger.info(f'All available bibtex entries have been stored in the file {filename_bibtex}')
        except Exception as e:
            logger.error(e)
            logger.error(f'A problem occurred when trying to write into the file {filename_bibtex}')
    if clipboard:
        import pyperclip
        try:
            pyperclip.copy(text)
            logger.info(f'All available bibtex entries have been stored in the system clipboard')
        except Exception as e:
            logger.error(e)
            logger.error(f'A problem occurred when trying to write into the system clipboard')

    
def main():
    parser = argparse.ArgumentParser( 
            description = "Generate BibTeX entries of scientific publications starting from the pdf files. It requires an internet connection.",
            epilog = "")

    parser.add_argument("path",
                        help = "Relative path of the target pdf file or of the targe folder.",
                        metavar = "path",
                        nargs = '*')
    parser.add_argument("-nv",
                        "--no_verbose",
                        help="Decrease verbosity.",
                        action="store_true")
    parser.add_argument("-s",
                        "--make_bibtex_file",
                        dest="filename_bibtex",
                        help="Create a text file inside the target directory, with name given by FILENAME_BIBTEX, containing the bibtex entry of each pdf file in the target folder (if any is found).",
                        action="store")
    parser.add_argument("-bclip",
                        "--save_bibtex_clipboard",
                        action="store_true",
                        help="Store all found bibtex entries into the clipboard.")
    parser.add_argument("-install--right--click",
                        dest="install_right_click",
                        action="store_true",
                        help="Add a shortcut to pdf2bib in the right-click context menu of Windows. You can copy the bibtex entry of a pdf file (or all pdf files in a folder) into the clipboard by just right clicking on it! NOTE: this feature is only available on Windows.")
    parser.add_argument("-uninstall--right--click",
                        dest="uninstall_right_click",
                        action="store_true",
                        help="Uninstall the right-click context menu functionalities. NOTE: this feature is only available on Windows.")

    args = parser.parse_args()

    # Setup logging
    if not(args.no_verbose): loglevel = logging.INFO
    else: loglevel = logging.CRITICAL

    logger = logging.getLogger("pdf2bib")
    logger.setLevel(level=loglevel)

    #If the command -install--right--click was specified, it sets the right keys in the system registry
    if args.install_right_click:
        import pdf2bib.utils_registry as utils_registry
        utils_registry.install_right_click()
        return
    if args.uninstall_right_click:
        import pdf2bib.utils_registry as utils_registry
        utils_registry.uninstall_right_click()
        return
    if isinstance(args.path,list):
        if len(args.path)>0:
            target = args.path[0]
        else:
            target = ""
    else:
        target = args.path

    if target == "":
        print("pdf2bib: error: the following arguments are required: path. Type \'pdf2bib --h\' for a list of commands.")
        return

    config.set('verbose',not(args.no_verbose))
 
    results = pdf2bib(target=target)

    if not results:
        return
    if not isinstance(results,list):
        results = [results]
    print('The following files were analyzed:')
    for result in results:
        if result['identifier']:
            print('{:<15s} {:<40s} {:<10s}\n'.format(result['identifier_type'], result['identifier'],result['path']) ) 
        else:
            print('{:<15s} {:<40s} {:<10s}\n'.format('n.a.', 'n.a.',result['path']) ) 

    # We call the function save_bibtex_entries. If args.filename_bibtex is a valid string, it will save all found identifiers in a text file with that name.
    # If args.save_doi_clipboard is true, it will copy all identifiers into the clipboard
    save_bibtex_entries(args.filename_bibtex, results, args.save_bibtex_clipboard)  

    return

if __name__ == '__main__':
    main()


import requests
import json
import feedparser
import re
import logging
import urllib.parse
from unidecode import unidecode
import bibtexparser
import pdf2doi.config as config

logger = logging.getLogger('pdf2bib')

#The functions parse_bib_from_dxdoiorg and parse_bib_from_exportarxivorg take as input a string or other object, which has been obtained by the library pdf2doi by
#querying the relevant websites. By analyzing this input, they create a dictionary containing valid bibtex infos and return it as output.
#The output dictionaries returned by parse_bib_from_dxdoiorg and parse_bib_from_exportarxivorg must have the same "format", i.e. the same set of required keys
#(additional keys are possible)
#The required keys in the dictionaries are 'title', 'journal', 'volume', 'issue', 'page', 'year', 'month', 'publisher', 'author', 'URL', 'DOI'    

#The dictionaries are then fed to the function make_bibtex(data) which creates a string containing the full bibtex entry

#The functions make_bibtex, parse_bib_from_dxdoiorg and parse_bib_from_exportarxivorg are called directly from the function pdf2bib_singlefile inside main.py, but they
#can also called directly by an user is (s)he knows what (s)he is doing.

def parse_bib_from_dxdoiorg(text, method):
    """
    Given a certain string contained in the input variable text, which was obtained from pdf2doi by quering dx.doi.org, it parses the string text and return a dictionary containing valid bibtex infos.
    The input variable method is a string which specify the method used by pdf2doi to query dx.doi.org, and it is normally defined in pdf2doi.get('method_dxdoiorg')
    """

    #Note: the methods "application/x-bibtex" and "text/bibliography; style=bibtex" return strings in the same format
    #However, as of 2021 Nov 6, the method "application/x-bibtex" does not return the paper journal (probably due to a bug in dx.doi.org
    #Note: "text/bibliography; style=bibtex" returns the authors in the format "LastName1, FirstName1 SecondName1.. and LastName2, FirstName2 SecondName2.. and etc."
    #which is not the format expect by the script pdf-renamer
    #Note: the method 'application/citeproc+json' returns data in JSON format
    # 
    if method == "application/x-bibtex":
        data = bibtexparser.loads(text)
        metadata = data.entries[0]
        return metadata
    if method == "text/bibliography; style=bibtex":
        data = bibtexparser.loads(text)
        metadata = data.entries[0]
        return metadata
    if method == "application/citeproc+json":
        json_dict = json.loads(text)
        #I extract only certain fields from the JSON dict
        fields = ['title', 'volume', 'issue', 'page', 'publisher', 'URL', 'DOI']
        metadata = dict()
        for field in fields:
            try:
                metadata[field.lower()] = json_dict[field] if field in json_dict.keys() else ''
            except:
                metadata[field.lower()] = ''

        try:
            metadata['journal'] = json_dict["container-title"] 
        except:
            metadata['journal'] = ''

        try:
            metadata['year'] = json_dict["issued"]['date-parts'][0][0]
        except:
            metadata['year'] = ''

        try:
            metadata['month'] = json_dict["issued"]['date-parts'][0][1]
        except:
            metadata['month'] = ''
        try:
            metadata['author'] = [author['given'] + ' ' + author['family'] for author in json_dict['author']]
        except:
            metadata['author'] = ''
        return metadata
    raise ValueError("The input variable method does not have a valid value")


def parse_bib_from_exportarxivorg(items):
    """
    Given a certain dictionary contained in the input variable items, which was obtained from pdf2doi by quering export.arxiv.org,
    it returns a dictionary containing valid bibtex infos
    """
    #Extract data from the dictionary items
    data_to_extract = ['title','authors','author','link','published','arxiv_doi']
    data =[items[key] if key in items.keys() else None for key in data_to_extract]

    #Create the dictionary data_dict which will be passed to the function make_bibtex
    data_dict = dict(zip(data_to_extract,data))

    #add additionaly values
    data_dict['eprint'] ="arXiv:" + arxivID 
    data_dict['ejournal'] ="arXiv" 
    data_dict['ENTRYTYPE'] = 'article'
    #rename some of the keys in order to match the bibtex standards
    data_dict['url'] = data_dict.pop('link')
    data_dict['doi'] = data_dict.pop('arxiv_doi')       
    #parse the published data to get the year, month and day
    if data_dict['published']:
        regexDate = re.search('(\d{4}\-\d{2}\-\d{2})',data_dict['published'],re.I)
        if regexDate:
            date_list =  (regexDate.group(1)).split("-")
            year = date_list[0] if len(date_list)>0 else '0000'
            month = date_list[1] if len(date_list)>1 else '00'
            day = date_list[2] if len(date_list)>2 else '00'
    else:
        year,month,day = '0000', '00', '00'
    data_dict['year'] = year
    data_dict['month'] = month
    data_dict['day'] = day

    if 'authors' in data_dict:
        authors = data_dict['authors']
    elif 'author' in data_dict:
        authors = data_dict['author']
    else:
        authors = ''
        
    #if authors are defined as list, create a string out of it, with the format 
    #"Name1 Lastname1 and Name2 Lastname2 and ... "
    if authors and isinstance(authors,list):
        authorsnames_list = [author['name'] for author in authors]
        data_dict['authors'] = " and ".join(authorsnames_list)

    return data_dict


def make_bibtex(data):
    #Based on the metadata contained in the input dictionary data, it creates a valid bibtex entry
    #The ID of the bibtex entry has the format [lastname_firstauthor][year][first_word_title] all in lower case
    #If the tag url is present, any possible ascii code (e.g. %2f) is decoded
    #Note: the code below assumes that the field for the authors is either a string in the format "Name1 Lastname1 and Name2 Lastname2 and ... "
    #or a list in the format ['Name1 Lastname1','Name2 Lastname2',...]

    
    if 'authors' in data.keys():
        authors = data['authors']
    elif 'author' in data.keys():
        authors = data['author']
    else:
        authors = ''
    if  isinstance(authors,list):
        authors = " and ".join(authors)
        data['author'] = authors
    if not(type(authors) in (str, list)):
        raise TypeError('The value corresponding to the key ''author'' or ''authors'' must be either a string or a list of strings')

    #After this line, authors must be a string
    
    #Generate the ID by looking for last name of first author, year of publicaton, and first word of title
    if authors:
        firstauthor = authors.split(' and ')[0]
        lastname_firstauthor = (firstauthor.strip()).split(' ')[-1]
    else: 
        lastname_firstauthor = ''
    year = data['year'] if 'year' in data.keys() else ''
    first_word_title =  data['title'].split(' ')[0] if 'title' in data.keys() else ''
    id = lastname_firstauthor + str(year) + first_word_title
    id = id.lower()
    id = remove_latex_codes(id)
    id = unidecode(id) #This makes sure that the id of the bibtex entry is only made out of ascii characters (i.e. no accents, tildes, etc.)
    if id == '':
        id = 'NoValidID'

    #Sanitize the URL
    if 'url' in data.keys():
        data['url'] = urllib.parse.unquote(data['url'])

    if not 'ENTRYTYPE' in data.keys():
        data['ENTRYTYPE'] = 'article'

    #Create the bibtex entryr as a string 
    metadata_not_to_use = ['ENTRYTYPE','ID'] #These are temporary metadata, not useful for bibtex
    text = ["@"+data['ENTRYTYPE']+"{" + id]
    for key, value in data.items():
        if value and not (key in metadata_not_to_use):
            text.append("\t%s = {%s}" % (key, value))
    bibtex_entry = (",\n").join(text) + "\n" + "}"
    return bibtex_entry


def remove_latex_codes(text):
    #It replaces any latex special code (e.g. {\`{u}}) by the "closest" unicode character (e.g. u). This is useful when
    #certain strings which might contain latex codes need to be used in contexts where only unicode characters are accepted
    
    #This regex looks for any substring that matches the pattern "{\string1{string2}}" where string1 can be anything,
    #and it replaces the whole substring by string2
    text_sanitized = re.sub(r"{\\[^\{]+{([\w]+)}}", r"\1",text)
    return text_sanitized